///////////////////////////////////////////////////////////////////////////////////////
//  calibration routines

/******* Adjust *******/
#pragma region Adjust
void setMaxResistanceMode() {
  // Sets DACs to 0. Adjust trimmers to measure 500K (when hot)...1M (when cold) on each LDR
  state = STATE_CALIB;
  setCalibrationRelays(OFF);
  PRINTLN("Adjust bias trimmers");
  setLSE(0); setLSH(0); setRSE(0); setRSH(0);

  oled.setCursor(0, 14);
  oled.clear();
  oled.setFont(textFont);
  oled.println("Adjust bias trimmers");
  oled.println("to 700K on LDRs");
  oled.println("and re-calibrate.");
  oled.println("Press btn to end.");

}

void setMinResistanceMode() {
  // Sets DACs to maximum. Measure the value of the LDR resistors and manually update the #define LDR_XXX_MIN above.
  state = STATE_CALIB;
  setCalibrationRelays(OFF);
  PRINTLN(msgMeas);
  setLSE_Range(LOW); setLSH_Range(LOW); setRSE_Range(LOW); setRSH_Range(LOW);
  setLSE(255); setLSH(255); setRSE(255); setRSH(255);


  oled.setCursor(0, 14);
  oled.clear();
  oled.setFont(textFont);
  oled.println("Measure min LDR vals");
  oled.println("then update the code");
  oled.println("and re-calibrate.");
  oled.println("Press btn to end.");


}
#pragma endregion


/******* Self-test *******/
#pragma region Self-test
byte doSelfTest() {
  //switch calibration relays on to measure LDRs
  setCalibrationRelays(ON);
  PRINTLN(msgTest);

  oled.clear();
  oled.setCursor(20, 3);
  oled.setFont(runFont);
  oled.print(msgTest);


  // test if min/max resistance values are within limits
  // test series LDRs max val
  setLSE_Range(HIGH); setRSE_Range(HIGH); setLSE(0); setRSE(0);
  setLSH_Range(LOW); setRSH_Range(LOW); setLSH(255); setRSH(255); //set shunt R to minimum
  delay(3000);

  oled.print(".");


  unsigned long val;
  val = getRLSE();
  if (val < LDR_ZERO_MIN) {
    PRINT("Error 10 "); PRINTLN(val);
    return 10;
  }
  val = getRRSE();
  if (val < LDR_ZERO_MIN) {
    PRINTLN("Error 11 "); PRINTLN(val);
    return 11;
  }


  // test shunt LDRs max val
  setLSE_Range(LOW); setRSE_Range(LOW); setLSE(255); setRSE(255); //set series R to minimum
  setLSH_Range(HIGH); setRSH_Range(HIGH); setLSH(0); setRSH(0);
  delay(4000);

  oled.print(".");


  val = getRLSH();
  if (val < LDR_ZERO_MIN) {
    PRINTLN("Error 12 "); PRINTLN(val);
    return 12;
  }
  val = getRRSH();
  if (val < LDR_ZERO_MIN) {
    PRINTLN("Error 13 "); PRINTLN(val);
    return 13;
  }

  return 0;
}
#pragma endregion


/******* Calibration *******/

#pragma region Calibration
byte doCalibration() {

  calibStarted = millis();
  //** local vars
  float att;
  unsigned long LSE_target, LSH_target, RSE_target, RSH_target, targetL0, targetR0;
  byte pwL0, pwL1, pwR0, pwR1 = 0;
  word iL0, iL1, iR0, iR1 = 0;
  unsigned long rL0, rL1, rR0, rR1 = 0;
  byte rangeL, rangeR;
  byte dpw, drm;
  bool foundL, foundR = false;
  bool reloopL, reloopR = false;
  bool nojumpL, nojumpR = false;

  //  setLCDMaxLight();
  state = STATE_CALIB;

  errc = doSelfTest();
  if (errc != 0)
    return errc;

  PRINTLN("Calibration");

  tick = 0;
  percent = 0;

  setCalibrationRelays(ON);

  //** get minimum R in HIGH R range
  setLSE_Range(HIGH); setRSE_Range(HIGH); setLSE(255); setRSE(255); //set series R to High range, minimum value
  setLSH_Range(LOW); setRSH_Range(LOW); setLSH(255); setRSH(255); //set shunt R to minimum
  delay(2000);
  unsigned long LSE_Min_Hi = getRLSE();
  unsigned long RSE_Min_Hi = getRRSE();
  oled.clear();
  printTick();

  setLSE_Range(LOW); setRSE_Range(LOW); setLSE(255); setRSE(255); //set series R to minimum
  setLSH_Range(HIGH); setRSH_Range(HIGH); setLSH(250); setRSH(250); //set shunt R to High range, almost minimum value
  delay(2000);
  unsigned long LSH_Min_Hi = getRLSH();
  unsigned long RSH_Min_Hi = getRRSH();
  setLSE(0); setLSH(0); setRSE(0); setRSH(0);

  printTick();

  delay(2000);

  printTick();

  PRINT(LSH_Min_Hi); PRINT("   "); PRINTLN(RSH_Min_Hi);



  //** search for series R current
  #pragma region series
  //series: the resistances decrease. start from maximum value and go down.
  rL0 = rR0 = 4000000;
  pwL0 = pwR0 = 0; pwL1 = pwR1 = 1;
  iL0 = iR0 = 0;
  setLSE_Range(HIGH); setRSE_Range(HIGH); setLSE(0); setRSE(0);
  setLSH_Range(LOW); setRSH_Range(LOW); setLSH(255); setRSH(255);


  //loop for each volume sample needed
  for (byte i = 0; i <= VOL_MAX_STEP - 3; i++) {
    //byte step = i+2;



    //** calculate attenuation needed
    att = getAttFromStep(i + 2);

    //** calculate target R values
    LSE_target = RSE_target = getRxFromAttAndImp(att, NOM_IMPEDANCE);
    LSH_target = RSH_target = getRyFromAttAndImp(att, NOM_IMPEDANCE);

    //** adjust shunt R for the parallel load impedance
    LSH_target = LOAD_IMPEDANCE * LSH_target / (LOAD_IMPEDANCE - LSH_target);
    RSH_target = LOAD_IMPEDANCE * RSH_target / (LOAD_IMPEDANCE - RSH_target);

    //** adjust values if out of attainable range
    if (LSE_target < LSE_Min_Hi - LSE_Min_Hi * LDR_LOW_TOLERANCE / 100) {
      LSE_target = LSE_Min_Hi;
      LSH_target = LOAD_IMPEDANCE * getRyFromAttAndRx(att, LSE_target) / (LOAD_IMPEDANCE - getRyFromAttAndRx(att, LSE_target));
    }

    if (LSH_target < LDR_LSH_MIN - LDR_LSH_MIN * LDR_LOW_TOLERANCE / 100) {
      LSH_target = LDR_LSH_MIN;
      LSE_target = getRxFromAttAndRy(att, LSH_target);
    }

    if (RSE_target < RSE_Min_Hi - RSE_Min_Hi * LDR_LOW_TOLERANCE / 100) {
      RSE_target = RSE_Min_Hi;
      RSH_target = LOAD_IMPEDANCE * getRyFromAttAndRx(att, RSE_target) / (LOAD_IMPEDANCE - getRyFromAttAndRx(att, RSE_target));
    }

    if (RSH_target < LDR_RSH_MIN - LDR_RSH_MIN * LDR_LOW_TOLERANCE / 100) {
      RSH_target = LDR_RSH_MIN;
      RSE_target = getRxFromAttAndRy(att, RSH_target);
    }

    PRINT("** Step "); PRINT(i + 2);
    PRINT(" Att= "); PRINT(att);
    PRINT(" LSE_t= "); PRINT(LSE_target); //print the calculated R values
    PRINT(" LSH_t= "); PRINT(LSH_target);
    PRINT(" RSE_t= "); PRINT(RSE_target);
    PRINT(" RSH_t= "); PRINTLN(RSH_target);




    //** find LDR current for the target resistances calculated above
    foundL = foundR = false;

    do {
      reloopL = reloopR = false;  //reloop = if the jump was too large, loop again with step=1

      setLSE(pwL1); setRSE(pwR1);

      if (digitalRead(PIN_BTN) == LOW)
        return 100; // abort on BTN press

      delay(500); // for the DAC to stabilize
      int del = getDelayHi(max(getILSE(), getIRSE())) * 100 - 500; // +variable delay
      PRINT("del: "); PRINTLN(del);
      if (del > 0) delay(del);
      printTick();

      if (!foundL) {
        rL1 = getRLSE();
        iL1 = getILSE();
      }
      if (!foundR) {
        rR1 = getRRSE();
        iR1 = getIRSE();
      }

      PRINTLN();
      if (!foundL) {
        PRINT("LSE: "); PRINT(pwL0);
        PRINT(" -> "); PRINT(rL0);
        PRINT("("); PRINT(iL0);
        PRINT(") - "); PRINT(LSE_target); PRINT(" - ");
        PRINT(pwL1); PRINT(" -> "); PRINT(rL1);
        PRINT("("); PRINT(iL1); PRINTLN(")");
      }
      if (!foundR) {
        PRINT("RSE: "); PRINT(pwR0);
        PRINT(" -> "); PRINT(rR0);
        PRINT("("); PRINT(iR0);
        PRINT(") - "); PRINT(RSE_target); PRINT(" - ");
        PRINT(pwR1); PRINT(" -> "); PRINT(rR1);
        PRINT("("); PRINT(iR1); PRINTLN(")");
      }



      // test if found L
      if (rL0 >= LSE_target && LSE_target >= rL1 && !foundL) {
        if (pwL1 - pwL0 > 1) {
          pwL1 = pwL0 + 1;
          reloopL = true; // found, but the PWM jump was too large, repeat with jump = 1
        }
        else {
          foundL = true;
          if (rL0 - LSE_target >= LSE_target - rL1) {
            dataL[i].pw_SE = pwL1;
            dataL[i].i_SE = iL1;
          }
          else
          {
            dataL[i].pw_SE = pwL0;
            dataL[i].i_SE = iL0;
          }
          PRINT("> Found LSE i="); PRINT(dataL[i].i_SE); PRINTLN();
        }
      }
      if (!foundL && !reloopL)
      {
        if (pwL1 == 255) {
          PRINTLN("Error: LSE not found");
          return 1;
        }

        if (rL0 < LSE_target && i > 1) { //go back
          PRINTLN("Err: OOB LSE");
          pwL0--; pwL1--;
          rL0 = 4000000;
        }
        else
        {
          if (rL0 != rL1 && rL0 != 4000000)
            dpw = abs((rL1 - LSE_target) * (pwL1 - pwL0) / (rL0 - rL1));
          else
            dpw = 1;
          if (dpw == 0) dpw = 1;
          pwL0 = pwL1;
          pwL1 += dpw;


          iL0 = iL1;
          rL0 = rL1;
        }
      }


      // test if found R
      if (rR0 >= RSE_target && RSE_target >= rR1 && !foundR) {
        if (pwR1 - pwR0 > 1) {
          pwR1 = pwR0 + 1;
          reloopR = true;
        }
        else {
          foundR = true;
          if (rR0 - RSE_target >= RSE_target - rR1) {
            dataR[i].pw_SE = pwR1;
            dataR[i].i_SE = iR1;
          }
          else
          {
            dataR[i].pw_SE = pwR0;
            dataR[i].i_SE = iR0;
          }
          PRINT("> Found RSE i="); PRINT(dataR[i].i_SE); PRINTLN();
        }
      }
      if (!foundR && !reloopR)
      {
        if (pwR1 == 255) {
          PRINTLN("Err: RSE not found");
          return 2;
        }

        if (rR0 < RSE_target && i > 1) { //go back
          PRINTLN("Err: OOB RSE");
          pwR0--; pwR1--;
          rR0 = 4000000;
        }
        else
        {
          if (rR0 != rR1 && rR0 != 4000000)
            dpw = (rR1 - RSE_target) * (pwR1 - pwR0) / (rR0 - rR1);
          else
            dpw = 1;
          if (dpw == 0) dpw = 1;
          pwR0 = pwR1;
          pwR1 += dpw;

          iR0 = iR1;
          rR0 = rR1;
        }
      }

    } while (!foundL || !foundR);




    percent = (i + 1) * 100 / (2 * VOL_MAX_STEP - 4);
    printTick();
    PRINT(percent); PRINTLN("%");




  } //end for series
  #pragma endregion series




  //** search for shunt R current
  #pragma region shunt
  //shunt: start at min R and go up
  rL0 = rR0 = 0;
  pwL0 = pwR0 = pwL1 = pwR1 = 255;
  iL0 = iR0 = 0;
  targetL0 = targetR0 = 0;
  rangeL = rangeR = LOW;
  setLSE_Range(LOW); setRSE_Range(LOW); setLSE(255); setRSE(255);
  setLSH_Range(LOW); setRSH_Range(LOW); setLSH(255); setRSH(255);
  delay(2000);

  printTick();

  //loop for each volume sample needed
  for (byte i = 0; i <= VOL_MAX_STEP - 3; i++) {
    //byte step = i+2;

    PRINTLN(); PRINT("** Step "); PRINTLN(i + 2);

    //** calculate target R values
    att = getAttFromStep(i + 2);

    LSE_target = RSE_target = getRxFromAttAndImp(att, NOM_IMPEDANCE);
    LSH_target = RSH_target = getRyFromAttAndImp(att, NOM_IMPEDANCE);

    //** adjust shunt R for the parallel load impedance
    LSH_target = LOAD_IMPEDANCE * LSH_target / (LOAD_IMPEDANCE - LSH_target);
    RSH_target = LOAD_IMPEDANCE * RSH_target / (LOAD_IMPEDANCE - RSH_target);

    if (LSE_target < LSE_Min_Hi - LSE_Min_Hi * LDR_LOW_TOLERANCE / 100) {
      LSE_target = LSE_Min_Hi;
      LSH_target = LOAD_IMPEDANCE * getRyFromAttAndRx(att, LSE_target) / (LOAD_IMPEDANCE - getRyFromAttAndRx(att, LSE_target));
    }

    if (LSH_target < LDR_LSH_MIN - LDR_LSH_MIN * LDR_LOW_TOLERANCE / 100) {
      LSH_target = LDR_LSH_MIN;
      LSE_target = getRxFromAttAndRy(att, LSH_target);
    }

    if (RSE_target < RSE_Min_Hi - RSE_Min_Hi * LDR_LOW_TOLERANCE / 100) {
      RSE_target = RSE_Min_Hi;
      RSH_target = LOAD_IMPEDANCE * getRyFromAttAndRx(att, RSE_target) / (LOAD_IMPEDANCE - getRyFromAttAndRx(att, RSE_target));
    }

    if (RSH_target < LDR_RSH_MIN - LDR_RSH_MIN * LDR_LOW_TOLERANCE / 100) {
      RSH_target = LDR_RSH_MIN;
      RSE_target = getRxFromAttAndRy(att, RSH_target);
    }


    //** find LDR current for the target resistances
    foundL = foundR = false;
    nojumpL = nojumpR = false;

    //switch range
    if (rangeL == LOW && (LSH_target > LSH_Min_Hi)) {
      PRINTLN("---> LSH HI range");
      rangeL = HIGH;
      setLSH_Range(HIGH);
      setLSH(255);
      delay(1000);
      rL0 = 0; iL0 = 0;
      pwL0 = pwL1 = 255;
      targetL0 = 0;
    }
    if (rangeR == LOW && (RSH_target > RSH_Min_Hi)) {
      PRINTLN("---> RSH HI range");
      rangeR = HIGH;
      setRSH_Range(HIGH);
      setRSH(255);
      delay(1000);
      rR0 = 0; iR0 = 0;
      pwR0 = pwR1 = 255;
      targetR0 = 0;
    }

    // already found the value?
    if (targetL0 == LSH_target) {
      dataL[i].pw_SH = dataL[i - 1].pw_SH;
      dataL[i].i_SH = dataL[i - 1].i_SH;
      PRINT("> Copy LSH i="); PRINT(dataL[i].i_SH); PRINTLN();
      foundL = true;
    }
    if (targetR0 == RSH_target) {
      dataR[i].pw_SH = dataR[i - 1].pw_SH;
      dataR[i].i_SH = dataR[i - 1].i_SH;
      PRINT("> Copy RSH i="); PRINT(dataL[i].i_SH); PRINTLN();
      foundR = true;
    }

    // search
    do {
      reloopL = reloopR = false;  //reloop = if the jump was too large, loop again with step=1
      setLSH(pwL1); setRSH(pwR1);

      if (digitalRead(PIN_BTN) == LOW)
        return 100; // abort on BTN press

      delay(500);
      if (rangeL == HIGH || rangeR == HIGH) {
        int del = getDelayHi(max(getILSH(), getIRSH())) * 100 - 500;
        PRINT("del "); PRINTLN(del);
        if (del > 0) delay(del);
      }
      printTick();

      if (!foundL) {
        rL1 = getRLSH();
        iL1 = getILSH();
      }
      if (!foundR) {
        rR1 = getRRSH();
        iR1 = getIRSH();
      }

      //      PRINTLN();
      if (!foundL) {
        PRINT("LSH: ");
        PRINT(pwL0); PRINT("->"); PRINT(rL0); PRINT("("); PRINT(iL0); PRINT(") - ");
        PRINT(LSH_target); PRINT(" - ");
        PRINT(pwL1); PRINT("->"); PRINT(rL1); PRINT("("); PRINT(iL1); PRINTLN(")");
      }
      if (!foundR) {
        PRINT("RSH: ");
        PRINT(pwR0); PRINT("->"); PRINT(rR0); PRINT("("); PRINT(iR0); PRINT(") - ");
        PRINT(RSH_target); PRINT(" - ");
        PRINT(pwR1); PRINT("->"); PRINT(rR1); PRINT("("); PRINT(iR1); PRINTLN(")");
      }


      // test if found L
      if (rL0 <= LSH_target && LSH_target <= rL1 && !foundL) {
        if (pwL0 - pwL1 > 1) {
          pwL1 = pwL0 - 1;
          nojumpL = reloopL = true;
        }
        else {
          foundL = true;
          if (LSH_target - rL0 < rL1 - LSH_target) {
            dataL[i].pw_SH = pwL0;
            dataL[i].i_SH = iL0;
          }
          else {
            dataL[i].pw_SH = pwL1;
            dataL[i].i_SH = iL1;
          }
          targetL0 = LSH_target;
          PRINT("> Found LSH i="); PRINT(dataL[i].i_SH); PRINTLN();
          if (rangeL == LOW)
            dataL[i].i_SH = dataL[i].i_SH | 32768; //bit15 == 1 means LOW R range
        }
      }

      if (!foundL && !reloopL) {
        if (pwL1 == 0) {
          PRINTLN("Err: LSH not found");
          return 3;
        }

        else if (rL0 > LSH_target && pwL0 < 255) { //go back
          PRINTLN("Err: OOB LSH");
          pwL0++; pwL1++;
          rL0 = 0;
        }

        else {
          if (pwL0 - pwL1 == 1 && rL1 - rL0 > 2 && !nojumpL) {
            if (rL0 != rL1 && rL0 != 0)
              dpw = abs((LSH_target - rL1) / (rL1 - rL0) / 2);
            else
              dpw = 1;
            if (dpw == 0) dpw = 1;
          }
          else dpw = 1;
          pwL0 = pwL1;
          pwL1 -= dpw;

          iL0 = iL1;
          rL0 = rL1;
        }
      }


      // test if found R
      if (rR0 <= RSH_target && RSH_target <= rR1 && !foundR) {
        if (pwR0 - pwR1 > 1) {
          pwR1 = pwR0 - 1;
          nojumpR = reloopR = true;
        }
        else {
          foundR = true;
          if (RSH_target - rR0 < rR1 - RSH_target) {
            dataR[i].pw_SH = pwR0;
            dataR[i].i_SH = iR0;
          }
          else {
            dataR[i].pw_SH = pwR1;
            dataR[i].i_SH = iR1;
          }
          targetR0 = RSH_target;
          PRINT("> Found RSH i="); PRINT(dataR[i].i_SH); PRINTLN();
          if (rangeR == LOW)
            dataR[i].i_SH = dataR[i].i_SH | 32768;
        }
      }

      if (!foundR && !reloopR) {
        if (pwR1 == 0) {
          PRINTLN("Err: RSH not found");
          return 4;
        }

        else if (rR0 > RSH_target && pwR0 < 255) { //go back
          PRINTLN("Err: OOB RSH");
          pwR0++; pwR1++;
          rR0 = 0;
        }

        else
        {
          if (pwR0 - pwR1 == 1 && rR1 - rR0 > 2 && !nojumpR) {
            if (rR0 != rR1 && rR0 != 0)
              dpw = abs((RSH_target - rR1) / (rR1 - rR0) / 2);
            else
              dpw = 1;
            if (dpw == 0) dpw = 1;
          }
          else dpw = 1;
          pwR0 = pwR1;
          pwR1 -= dpw;


          iR0 = iR1;
          rR0 = rR1;
        }
      }

    } while (!foundL || !foundR);
    percent = (i + VOL_MAX_STEP - 1) * 100 / (2 * VOL_MAX_STEP - 4);
    printTick();
    PRINT(percent); PRINTLN("%");

  } //end for shunt
  #pragma endregion shunt

  volume = VOL_DEFAULT;
  saveCalibration();
  oled.clear();
  toRunState();
  drawRunDisplay();    // redraw full screen again
  setVolume(volume);
  setCalibrationRelays(OFF);

  return 0;
}
#pragma endregion
