//////////////////////////////////////////////////////////////////////////////////////
// LDR management
#pragma region LDR

/******* get attenuation factor from step number, LOG law *******/
float getAttFromStep(byte step) {
  float p = step / float(VOL_MAX_STEP);
  float divider = 100.0;
  if (p < 0.4)
    divider = 100.0 + pow((0.4 - p) * 50, 2.1);
  if (step == 2) divider = 1000.0;
  if (step == 3) divider = 600.0;
  return pow(10, p * 2) / divider;
}


/******* read LDR resistor value *******/
unsigned long getResistance(word measure, word seriesLDRres, word refResistor) {
  if (measure >= 4088)
    return 4000000;
  else {
    float v = LDR_VOLTAGE * measure / 4092.0;
    float result = v * refResistor / (LDR_VOLTAGE - v) - seriesLDRres;
    return long(result) + 25;
  }
}


/******* get oversampled 12-bit reading *******/
word getSample(byte pin) {
  word result = 0;
  for (byte i = 0; i < 16; i++) {
    result += analogRead(pin);
    delay(1);
  }
  return result >> 2;
}


/******* get LDR current *******/
word getILSE() {
  return getSample(PIN_SENSE_LSE);
}
word getILSH() {
  return getSample(PIN_SENSE_LSH);
}
word getIRSE() {
  return getSample(PIN_SENSE_RSE);
}
word getIRSH() {
  return getSample(PIN_SENSE_RSH);
}


/******* get LDR resistance *******/
unsigned long getRLSE() {
  return getResistance(getSample(PIN_LDR_L), LDR_LSH_MIN, LDR_R1);
}
unsigned long getRLSH() {
  return getResistance(getSample(PIN_LDR_L), LDR_LSE_MIN, LDR_R1);
}
unsigned long getRRSE() {
  return getResistance(getSample(PIN_LDR_R), LDR_RSH_MIN, LDR_R18);
}
unsigned long getRRSH() {
  return getResistance(getSample(PIN_LDR_R), LDR_RSE_MIN, LDR_R18);
}


/******* get series R from attenuation and impedance *******/
unsigned long getRxFromAttAndImp(float att, unsigned long imp) {
  return long(imp - att * imp);
}

/******* get shunt R from attenuation and impedance *******/
unsigned long getRyFromAttAndImp(float att, unsigned long imp) {
  return long(att * imp);
}

/******* get series R from attenuation and shunt R *******/
unsigned long getRxFromAttAndRy(float att, unsigned long ry) {
  return long(ry * (1 - att) / att);
}

/******* get shunt R from attenuation and series R *******/
unsigned long getRyFromAttAndRx(float att, unsigned long rx) {
  return long(att * rx / (1 - att));
}


/******* set DAC PWM *******/
inline void setLSE(byte val) {
  analogWrite(PIN_DAC_LSE, val);
}
inline void setLSH(byte val) {
  analogWrite(PIN_DAC_LSH, val);
}
inline void setRSE(byte val) {
  analogWrite(PIN_DAC_RSE, val);
}
inline void setRSH(byte val) {
  analogWrite(PIN_DAC_RSH, val);
}


/******* set high current/low resistance bias range *******/
// LOW = low resistance range
inline void setLSE_Range(byte val) {
  pinMode(PIN_SENSE_LSE, val == LOW ? OUTPUT : INPUT);
}
inline void setRSE_Range(byte val) {
  pinMode(PIN_SENSE_RSE, val == LOW ? OUTPUT : INPUT);
}
inline void setLSH_Range(byte val) {
  pinMode(PIN_BIAS_LSH, val == LOW ? OUTPUT : INPUT);
  LSHrange = val;
}
inline void setRSH_Range(byte val) {
  pinMode(PIN_BIAS_RSH, val == LOW ? OUTPUT : INPUT);
  RSHrange = val;
}


/** average delay needed for LDR settling **/
byte getDelayHi(word i) {
  if (i < 136) return 100;
  if (i < 240) return 70;
  if (i < 290) return 60;
  if (i < 340) return 50;
  if (i < 390) return 40;
  if (i < 590) return 30;
  if (i < 700) return 25;
  if (i < 900) return 20;
  if (i < 1300) return 15;
  if (i < 2100) return 10;
  if (i < 2500) return 8;
  else return 5;
}

word getDelayLo(byte pw) {
  return map(pw, 0, 255, 500, 700);
}

/** average difference in I (current) samples when PWM is increased by 1 **/
byte getDeltaI(word sample) {
  word s = sample & 32767;
  if (sample < 32768) {
    if (s < 42) return 5;
    if (s < 57) return 6;
    if (s < 100) return 7;
    if (s < 150) return 8;
    if (s < 210) return 9;
    if (s < 500) return 10;
    else return 11;
  }
  else {
    if (s < 16) return 3;
    if (s < 32) return 4;
    if (s < 50) return 5;
    if (s < 70) return 6;
    if (s < 100) return 7;
    if (s < 180) return 8;
    if (s < 300) return 9;
    if (s < 640) return 10;
    else return 11;
  }
}

#pragma endregion
